import babelGenerator from '@babel/generator';
import { parse } from '@babel/parser';
import babelPluginProposalDecorators from '@babel/plugin-proposal-decorators';
import babelPluginProposalDestructuringPrivate from '@babel/plugin-proposal-destructuring-private';
import babelPluginProposalExplicitResourceManagement from '@babel/plugin-proposal-explicit-resource-management';
import babelPluginTransformClassProperties from '@babel/plugin-transform-class-properties';
import babelPluginTransformClassStaticBlock from '@babel/plugin-transform-class-static-block';
import babelPluginTransformExponentiationOperator from '@babel/plugin-transform-exponentiation-operator';
import babelPluginTransformJsonStrings from '@babel/plugin-transform-json-strings';
import babelPluginTransformLogicalAssignmentOperators from '@babel/plugin-transform-logical-assignment-operators';
import babelPluginTransformNullishCoalescingOperator from '@babel/plugin-transform-nullish-coalescing-operator';
import babelPluginTransformNumericSeparator from '@babel/plugin-transform-numeric-separator';
import babelPluginTransformObjectRestSpread from '@babel/plugin-transform-object-rest-spread';
import babelPluginTransformOptionalCatchBinding from '@babel/plugin-transform-optional-catch-binding';
import babelPluginTransformOptionalChaining from '@babel/plugin-transform-optional-chaining';
import babelPluginTransformPrivatePropertyInObject from '@babel/plugin-transform-private-property-in-object';
import babelPluginTransformUnicodeSetsRegex from '@babel/plugin-transform-unicode-sets-regex';
import babelTraverse from '@babel/traverse';
import t from '@babel/types';
import { babel } from '@rollup/plugin-babel';
import rollupPluginCommonJS from '@rollup/plugin-commonjs';
import rollupPluginJSON from '@rollup/plugin-json';
import rollupPluginNodeResolve from '@rollup/plugin-node-resolve';
import { assert } from '@samual/lib/assert';
import { resolve } from 'path';
import prettier from 'prettier';
import { rollup } from 'rollup';
import { supportedExtensions } from '../constants.js';
import { minify } from './minify.js';
import { postprocess } from './postprocess.js';
import { preprocess } from './preprocess.js';
import { includesIllegalString, replaceUnsafeStrings } from './shared.js';
import { transform } from './transform.js';
import '@samual/lib/countHackmudCharacters';
import '@samual/lib/spliceString';
import 'acorn';
import 'terser';
import 'import-meta-resolve';
import '@samual/lib/clearObject';

const {
  format
} = prettier;
// eslint-disable-next-line @typescript-eslint/consistent-type-imports
const {
  default: generate
} = babelGenerator;
// eslint-disable-next-line @typescript-eslint/consistent-type-imports
const {
  default: traverse
} = babelTraverse;
/**
 * Minifies a given script
 *
 * @param code JavaScript or TypeScript code
 * @param options {@link ProcessOptions details}
 */
const processScript = async (code, {
  minify: shouldMinify = true,
  uniqueID = Math.floor(Math.random() * 2 ** 52).toString(36).padStart(11, `0`),
  scriptUser = `UNKNOWN`,
  scriptName = `UNKNOWN`,
  filePath,
  mangleNames = false,
  forceQuineCheats
} = {}) => {
  assert(/^\w{11}$/.exec(uniqueID));
  const sourceCode = code;
  let autocomplete;
  let statedSeclevel;

  // TODO do seclevel detection and verification per module

  const autocompleteMatch = /^function\s*\(.+\/\/(?<autocomplete>.+)/.exec(code);
  if (autocompleteMatch) {
    code = `export default ${code}`;
    ({
      autocomplete
    } = autocompleteMatch.groups);
  } else {
    for (const line of code.split(`\n`)) {
      const comment = /^\s*\/\/(?<commentContent>.+)/.exec(line);
      if (!comment) break;
      const commentContent = comment.groups.commentContent.trim();
      if (commentContent.startsWith(`@autocomplete `)) autocomplete = commentContent.slice(14).trimStart();else if (commentContent.startsWith(`@seclevel `)) {
        const seclevelString = commentContent.slice(10).trimStart().toLowerCase();
        switch (seclevelString) {
          case `fullsec`:
          case `full`:
          case `fs`:
          case `4s`:
          case `f`:
          case `4`:
            {
              statedSeclevel = 4;
            }
            break;
          case `highsec`:
          case `high`:
          case `hs`:
          case `3s`:
          case `h`:
          case `3`:
            {
              statedSeclevel = 3;
            }
            break;
          case `midsec`:
          case `mid`:
          case `ms`:
          case `2s`:
          case `m`:
          case `2`:
            {
              statedSeclevel = 2;
            }
            break;
          case `lowsec`:
          case `low`:
          case `ls`:
          case `1s`:
          case `l`:
          case `1`:
            {
              statedSeclevel = 1;
            }
            break;
          case `nullsec`:
          case `null`:
          case `ns`:
          case `0s`:
          case `n`:
          case `0`:
            {
              statedSeclevel = 0;
            }
            break;
          default:
            // TODO turn into warning when I get round to those
            throw new Error(`unrecognised seclevel "${seclevelString}"`);
        }
      }
    }
  }
  assert(/^\w{11}$/.exec(uniqueID));
  const plugins = [[babelPluginProposalDecorators.default, {
    decoratorsBeforeExport: true
  }], [babelPluginTransformClassProperties.default], [babelPluginTransformClassStaticBlock.default], [babelPluginTransformPrivatePropertyInObject.default], [babelPluginTransformLogicalAssignmentOperators.default], [babelPluginTransformNumericSeparator.default], [babelPluginTransformNullishCoalescingOperator.default], [babelPluginTransformOptionalChaining.default], [babelPluginTransformOptionalCatchBinding.default], [babelPluginTransformJsonStrings.default], [babelPluginTransformObjectRestSpread.default], [babelPluginTransformExponentiationOperator.default], [babelPluginTransformUnicodeSetsRegex.default], [babelPluginProposalDestructuringPrivate.default], [babelPluginProposalExplicitResourceManagement.default]];
  let filePathResolved;
  if (filePath) {
    filePathResolved = resolve(filePath);
    if (filePath.endsWith(`.ts`)) plugins.push([(await import('@babel/plugin-transform-typescript')).default, {
      allowDeclareFields: true,
      optimizeConstEnums: true
    }]);else {
      const [babelPluginProposalDoExpressions, babelPluginProposalFunctionBind, babelPluginProposalFunctionSent, babelPluginProposalPartialApplication, babelPluginProposalPipelineOperator, babelPluginProposalThrowExpressions, babelPluginProposalRecordAndTuple] = await Promise.all([import('@babel/plugin-proposal-do-expressions'), import('@babel/plugin-proposal-function-bind'), import('@babel/plugin-proposal-function-sent'), import('@babel/plugin-proposal-partial-application'), import('@babel/plugin-proposal-pipeline-operator'), import('@babel/plugin-proposal-throw-expressions'), import('@babel/plugin-proposal-record-and-tuple')]);
      plugins.push([babelPluginProposalDoExpressions.default], [babelPluginProposalFunctionBind.default], [babelPluginProposalFunctionSent.default], [babelPluginProposalPartialApplication.default], [babelPluginProposalPipelineOperator.default, {
        proposal: `hack`,
        topicToken: `%`
      }], [babelPluginProposalThrowExpressions.default], [babelPluginProposalRecordAndTuple.default, {
        syntaxType: `hash`,
        importPolyfill: true
      }]);
    }
  } else {
    filePathResolved = `${uniqueID}.ts`;
    const [babelPluginTransformTypescript, babelPluginProposalDoExpressions, babelPluginProposalFunctionBind, babelPluginProposalFunctionSent, babelPluginProposalPartialApplication, babelPluginProposalPipelineOperator, babelPluginProposalThrowExpressions, babelPluginProposalRecordAndTuple] = await Promise.all([import('@babel/plugin-transform-typescript'), import('@babel/plugin-proposal-do-expressions'), import('@babel/plugin-proposal-function-bind'), import('@babel/plugin-proposal-function-sent'), import('@babel/plugin-proposal-partial-application'), import('@babel/plugin-proposal-pipeline-operator'), import('@babel/plugin-proposal-throw-expressions'), import('@babel/plugin-proposal-record-and-tuple')]);
    plugins.push([babelPluginTransformTypescript.default, {
      allowDeclareFields: true,
      optimizeConstEnums: true
    }], [babelPluginProposalDoExpressions.default], [babelPluginProposalFunctionBind.default], [babelPluginProposalFunctionSent.default], [babelPluginProposalPartialApplication.default], [babelPluginProposalPipelineOperator.default, {
      proposal: `hack`,
      topicToken: `%`
    }], [babelPluginProposalThrowExpressions.default], [babelPluginProposalRecordAndTuple.default, {
      syntaxType: `hash`,
      importPolyfill: true
    }]);
  }
  const bundle = await rollup({
    input: filePathResolved,
    plugins: [{
      name: `hackmud-script-manager`,
      transform: async code => (await preprocess(code, {
        uniqueID
      })).code
    }, babel({
      babelHelpers: `bundled`,
      plugins,
      configFile: false,
      extensions: supportedExtensions
    }), rollupPluginCommonJS(), rollupPluginNodeResolve({
      extensions: supportedExtensions
    }), rollupPluginJSON()],
    treeshake: {
      moduleSideEffects: false
    }
  });
  const seclevelNames = [`NULLSEC`, `LOWSEC`, `MIDSEC`, `HIGHSEC`, `FULLSEC`];
  code = (await bundle.generate({})).output[0].code;
  const {
    file,
    seclevel
  } = transform(parse(code, {
    sourceType: `module`
  }), sourceCode, {
    uniqueID,
    scriptUser,
    scriptName
  });
  if (statedSeclevel != undefined && seclevel < statedSeclevel)
    // TODO replace with a warning and build script anyway
    throw new Error(`detected seclevel ${seclevelNames[seclevel]} is lower than stated seclevel ${seclevelNames[statedSeclevel]}`);
  code = generate(file).code;
  if (shouldMinify) code = await minify(file, {
    uniqueID,
    mangleNames,
    forceQuineCheats,
    autocomplete
  });else {
    traverse(file, {
      MemberExpression({
        node: memberExpression
      }) {
        if (memberExpression.computed) return;
        assert(memberExpression.property.type == `Identifier`);
        if (memberExpression.property.name == `prototype`) {
          memberExpression.computed = true;
          memberExpression.property = t.stringLiteral(`prototype`);
        } else if (memberExpression.property.name == `__proto__`) {
          memberExpression.computed = true;
          memberExpression.property = t.stringLiteral(`__proto__`);
        } else if (includesIllegalString(memberExpression.property.name)) {
          memberExpression.computed = true;
          memberExpression.property = t.stringLiteral(replaceUnsafeStrings(uniqueID, memberExpression.property.name));
        }
      },
      VariableDeclarator(path) {
        const renameVariables = lValue => {
          switch (lValue.type) {
            case `Identifier`:
              {
                if (includesIllegalString(lValue.name)) path.scope.rename(lValue.name, `$${Math.floor(Math.random() * 2 ** 52).toString(36).padStart(11, `0`)}`);
              }
              break;
            case `ObjectPattern`:
              {
                for (const property of lValue.properties) {
                  assert(property.type == `ObjectProperty`);
                  renameVariables(property.value);
                }
              }
              break;
            case `ArrayPattern`:
              {
                for (const element of lValue.elements) {
                  if (element) renameVariables(element);
                }
              }
              break;
            default:
              throw new Error(`unknown lValue type "${lValue.type}"`);
          }
        };
        renameVariables(path.node.id);
      },
      ObjectProperty({
        node: objectProperty
      }) {
        if (objectProperty.key.type == `Identifier` && includesIllegalString(objectProperty.key.name)) {
          objectProperty.key = t.stringLiteral(replaceUnsafeStrings(uniqueID, objectProperty.key.name));
          objectProperty.shorthand = false;
        }
      },
      StringLiteral({
        node
      }) {
        node.value = replaceUnsafeStrings(uniqueID, node.value);
      },
      TemplateLiteral({
        node
      }) {
        for (const templateElement of node.quasis) {
          if (templateElement.value.cooked) {
            templateElement.value.cooked = replaceUnsafeStrings(uniqueID, templateElement.value.cooked);
            templateElement.value.raw = templateElement.value.cooked.replaceAll(`\\`, `\\\\`).replaceAll(`\``, `\\\``).replaceAll(`\${`, `$\\{`);
          } else templateElement.value.raw = replaceUnsafeStrings(uniqueID, templateElement.value.raw);
        }
      },
      RegExpLiteral(path) {
        path.node.pattern = replaceUnsafeStrings(uniqueID, path.node.pattern);
        delete path.node.extra;
      }
    });

    // we can't have comments because they may contain illegal strings
    code = await format(generate(file, {
      comments: false
    }).code, {
      parser: `babel`,
      arrowParens: `avoid`,
      semi: false,
      trailingComma: `none`
    });
  }
  code = postprocess(code, seclevel, uniqueID);
  if (includesIllegalString(code)) throw new Error(`you found a weird edge case where I wasn't able to replace illegal strings like "SC$", please report thx`);
  return {
    script: code,
    warnings: []
  };
};

export { processScript as default, minify, postprocess, preprocess, processScript, transform };
