import { DynamicMap } from '@samual/lib/DynamicMap';
import { countHackmudCharacters } from '@samual/lib/countHackmudCharacters';
import { writeFilePersistent } from '@samual/lib/writeFilePersistent';
import { readdir, readFile } from 'fs/promises';
import { resolve, extname, basename } from 'path';
import { supportedExtensions } from './constants.js';
import { processScript } from './processScript/index.js';
import '@babel/generator';
import '@babel/parser';
import '@babel/plugin-proposal-decorators';
import '@babel/plugin-proposal-destructuring-private';
import '@babel/plugin-proposal-explicit-resource-management';
import '@babel/plugin-transform-class-properties';
import '@babel/plugin-transform-class-static-block';
import '@babel/plugin-transform-exponentiation-operator';
import '@babel/plugin-transform-json-strings';
import '@babel/plugin-transform-logical-assignment-operators';
import '@babel/plugin-transform-nullish-coalescing-operator';
import '@babel/plugin-transform-numeric-separator';
import '@babel/plugin-transform-object-rest-spread';
import '@babel/plugin-transform-optional-catch-binding';
import '@babel/plugin-transform-optional-chaining';
import '@babel/plugin-transform-private-property-in-object';
import '@babel/plugin-transform-unicode-sets-regex';
import '@babel/traverse';
import '@babel/types';
import '@rollup/plugin-babel';
import '@rollup/plugin-commonjs';
import '@rollup/plugin-json';
import '@rollup/plugin-node-resolve';
import '@samual/lib/assert';
import 'prettier';
import 'rollup';
import './processScript/minify.js';
import '@samual/lib/spliceString';
import 'acorn';
import 'terser';
import './processScript/shared.js';
import './processScript/postprocess.js';
import './processScript/preprocess.js';
import 'import-meta-resolve';
import './processScript/transform.js';
import '@samual/lib/clearObject';

/**
 * Push scripts from a source directory to the hackmud directory.
 *
 * Pushes files directly in the source folder to all users
 * @param sourceDirectory directory containing source code
 * @param hackmudDirectory directory created by hackmud containing user data including scripts
 * @param options {@link PushOptions details}
 * @returns array of info on pushed scripts
 */
const push = async (sourceDirectory, hackmudDirectory, {
  scripts = [`*.*`],
  onPush = () => {},
  minify = true,
  mangleNames = false,
  forceQuineCheats
} = {}) => {
  const scriptNamesByUser = new DynamicMap(_user => new Set());
  const wildScriptUsers = new Set();
  const wildUserScripts = new Set();
  let pushEverything = false;
  for (const fullScriptName of scripts) {
    const [user, scriptName] = fullScriptName.split(`.`);
    if (!user || user == `*`) {
      if (!scriptName || scriptName == `*`) pushEverything = true;else wildUserScripts.add(scriptName);
    } else if (!scriptName || scriptName == `*`) wildScriptUsers.add(user);else scriptNamesByUser.get(user).add(scriptName);
  }
  const usersByGlobalScriptsToPush = new DynamicMap(_user => new Set());
  const allInfo = [];
  const scriptNamesAlreadyPushedByUser = new DynamicMap(_user => new Set());
  let sourceDirectoryDirents;

  // *.bar
  if (wildUserScripts.size || pushEverything) {
    const hackmudDirectoryDirents = await readdir(resolve(hackmudDirectory), {
      withFileTypes: true
    });
    const allUsers = new Set([...(sourceDirectoryDirents = await readdir(resolve(sourceDirectory), {
      withFileTypes: true
    })).filter(dirent => dirent.isDirectory()).map(dirent => dirent.name), ...hackmudDirectoryDirents.filter(dirent => dirent.isDirectory()).map(dirent => dirent.name), ...hackmudDirectoryDirents.filter(dirent => dirent.isFile() && dirent.name.endsWith(`.key`)).map(dirent => dirent.name.slice(0, -4)), ...scriptNamesByUser.keys(), ...wildScriptUsers]);
    if (pushEverything) {
      for (const user of allUsers) wildScriptUsers.add(user);
    } else {
      for (const user of allUsers) {
        const scriptNames = scriptNamesByUser.get(user);
        for (const scriptName of wildUserScripts) scriptNames.add(scriptName);
      }
    }
  }

  // foo.*
  await Promise.all([...wildScriptUsers].map(async user => {
    await readdir(resolve(sourceDirectory, user), {
      withFileTypes: true
    }).then(async dirents => {
      await Promise.all(dirents.map(async dirent => {
        if (dirent.name.endsWith(`.d.ts`)) return;
        const extension = extname(dirent.name);
        if (dirent.isFile() && supportedExtensions.includes(extension)) {
          const scriptName = basename(dirent.name, extension);
          const filePath = resolve(sourceDirectory, user, dirent.name);
          const {
            script: minifiedCode
          } = await processScript(await readFile(filePath, {
            encoding: `utf-8`
          }), {
            minify,
            scriptUser: user,
            scriptName,
            filePath,
            mangleNames,
            forceQuineCheats
          });
          const info = {
            file: `${user}/${dirent.name}`,
            users: [user],
            minLength: countHackmudCharacters(minifiedCode),
            error: undefined
          };
          scriptNamesAlreadyPushedByUser.get(user).add(scriptName);
          allInfo.push(info);
          await writeFilePersistent(resolve(hackmudDirectory, user, `scripts/${scriptName}.js`), minifiedCode);
          onPush(info);
        }
      }));
    }, error => {
      if (error.code != `ENOENT`) throw error;
    });
  }));

  // foo.bar
  await Promise.all([...scriptNamesByUser].map(async ([user, scripts]) => {
    if (wildScriptUsers.has(user)) return;
    await Promise.all([...scripts].map(async scriptName => {
      let code;
      let fileName;
      let filePath;

      // TODO there's definitly a better way to do this
      for (const extension of supportedExtensions) {
        try {
          fileName = `${scriptName}${extension}`;
          // eslint-disable-next-line no-await-in-loop -- I don't think paralelysing this is worth it
          code = await readFile(filePath = resolve(sourceDirectory, user, fileName), {
            encoding: `utf-8`
          });
          break;
        } catch {}
      }
      if (code) {
        const {
          script: minifiedCode
        } = await processScript(code, {
          minify,
          scriptUser: user,
          scriptName,
          filePath,
          mangleNames,
          forceQuineCheats
        });
        const info = {
          file: `${user}/${fileName}`,
          users: [user],
          minLength: countHackmudCharacters(minifiedCode),
          error: undefined
        };
        allInfo.push(info);
        await writeFilePersistent(resolve(hackmudDirectory, user, `scripts`, `${scriptName}.js`), minifiedCode);
        onPush(info);
      } else usersByGlobalScriptsToPush.get(scriptName).add(user);
    }));
  }));

  // foo.* (global)
  await (wildScriptUsers.size ? Promise.all((sourceDirectoryDirents || (await readdir(resolve(sourceDirectory), {
    withFileTypes: true
  }))).map(async dirent => {
    if (dirent.name.endsWith(`.d.ts`)) return;
    const extension = extname(dirent.name);
    if (!dirent.isFile() || !supportedExtensions.includes(extension)) return;
    const scriptName = basename(dirent.name, extension);
    const usersToPushTo = [...wildScriptUsers, ...usersByGlobalScriptsToPush.get(scriptName)].filter(user => !scriptNamesAlreadyPushedByUser.get(user).has(scriptName));
    if (!usersToPushTo.length) return;
    const uniqueID = Math.floor(Math.random() * 2 ** 52).toString(36).padStart(11, `0`);
    const filePath = resolve(sourceDirectory, dirent.name);
    const {
      script: minifiedCode
    } = await processScript(await readFile(filePath, {
      encoding: `utf-8`
    }), {
      minify,
      scriptUser: true,
      scriptName,
      uniqueID,
      filePath,
      mangleNames,
      forceQuineCheats
    });
    const info = {
      file: dirent.name,
      users: usersToPushTo,
      minLength: countHackmudCharacters(minifiedCode),
      error: undefined
    };
    await Promise.all(usersToPushTo.map(user => writeFilePersistent(resolve(hackmudDirectory, user, `scripts/${scriptName}.js`), minifiedCode.replace(new RegExp(`\\$${uniqueID}\\$SCRIPT_USER\\$`, `g`), user).replace(new RegExp(`\\$${uniqueID}\\$FULL_SCRIPT_NAME\\$`, `g`), `${user}.${scriptName}`))));
    allInfo.push(info);
    onPush(info);
  })) : Promise.all([...usersByGlobalScriptsToPush].map(async ([scriptName, users]) => {
    let code;
    let fileName;
    let filePath;

    // TODO there's definitly a better way to do this
    for (const extension of supportedExtensions) {
      try {
        fileName = `${scriptName}${extension}`;
        // eslint-disable-next-line no-await-in-loop -- I don't think paralelysing this is worth it
        code = await readFile(filePath = resolve(sourceDirectory, fileName), {
          encoding: `utf-8`
        });
        break;
      } catch {}
    }
    if (code) {
      const uniqueID = Math.floor(Math.random() * 2 ** 52).toString(36).padStart(11, `0`);
      const {
        script: minifiedCode
      } = await processScript(code, {
        minify,
        scriptUser: true,
        scriptName,
        uniqueID,
        filePath,
        mangleNames,
        forceQuineCheats
      });
      const info = {
        file: fileName,
        users: [...users],
        minLength: countHackmudCharacters(minifiedCode),
        error: undefined
      };
      await Promise.all([...users].map(user => writeFilePersistent(resolve(hackmudDirectory, user, `scripts/${scriptName}.js`), minifiedCode.replace(new RegExp(`\\$${uniqueID}\\$SCRIPT_USER\\$`, `g`), user).replace(new RegExp(`\\$${uniqueID}\\$FULL_SCRIPT_NAME\\$`, `g`), `${user}.${scriptName}`))));
      allInfo.push(info);
      onPush(info);
    }
  })));
  return allInfo;
};

export { push as default, push };
