import { DynamicMap } from '@samual/lib/DynamicMap';
import { assert } from '@samual/lib/assert';
import { countHackmudCharacters } from '@samual/lib/countHackmudCharacters';
import { writeFilePersistent } from '@samual/lib/writeFilePersistent';
import { watch as watch$1 } from 'chokidar';
import { readdir, readFile, writeFile } from 'fs/promises';
import { extname, basename, resolve } from 'path';
import { supportedExtensions } from './constants.js';
import { generateTypeDeclaration } from './generateTypeDeclaration.js';
import { processScript } from './processScript/index.js';
import '@babel/generator';
import '@babel/parser';
import '@babel/plugin-proposal-decorators';
import '@babel/plugin-proposal-destructuring-private';
import '@babel/plugin-proposal-explicit-resource-management';
import '@babel/plugin-transform-class-properties';
import '@babel/plugin-transform-class-static-block';
import '@babel/plugin-transform-exponentiation-operator';
import '@babel/plugin-transform-json-strings';
import '@babel/plugin-transform-logical-assignment-operators';
import '@babel/plugin-transform-nullish-coalescing-operator';
import '@babel/plugin-transform-numeric-separator';
import '@babel/plugin-transform-object-rest-spread';
import '@babel/plugin-transform-optional-catch-binding';
import '@babel/plugin-transform-optional-chaining';
import '@babel/plugin-transform-private-property-in-object';
import '@babel/plugin-transform-unicode-sets-regex';
import '@babel/traverse';
import '@babel/types';
import '@rollup/plugin-babel';
import '@rollup/plugin-commonjs';
import '@rollup/plugin-json';
import '@rollup/plugin-node-resolve';
import 'prettier';
import 'rollup';
import './processScript/minify.js';
import '@samual/lib/spliceString';
import 'acorn';
import 'terser';
import './processScript/shared.js';
import './processScript/postprocess.js';
import './processScript/preprocess.js';
import 'import-meta-resolve';
import './processScript/transform.js';
import '@samual/lib/clearObject';

/**
 * Watches target file or folder for updates and builds and pushes updated file.
 *
 * @param sourceDirectory path to folder containing source files
 * @param hackmudDirectory path to hackmud directory
 * @param options {@link WatchOptions details} and {@link PushOptions more details}
 */
const watch = async (sourceDirectory, hackmudDirectory, {
  scripts = [`*.*`],
  onPush,
  minify = true,
  mangleNames = false,
  typeDeclarationPath: typeDeclarationPath_,
  onReady,
  forceQuineCheats
} = {}) => {
  if (!scripts.length) throw new Error(`scripts option was an empty array`);
  const scriptNamesToUsers = new DynamicMap(_scriptName => new Set());
  const wildScriptUsers = new Set();
  const wildUserScripts = new Set();
  let pushEverything = false;
  for (const fullScriptName of scripts) {
    const [user, scriptName] = fullScriptName.split(`.`);
    if (!user || user == `*`) {
      if (!scriptName || scriptName == `*`) pushEverything = true;else wildUserScripts.add(scriptName);
    } else if (!scriptName || scriptName == `*`) wildScriptUsers.add(user);else scriptNamesToUsers.get(scriptName).add(user);
  }
  const watcher = watch$1([`*.ts`, `*.js`], {
    depth: 1,
    cwd: sourceDirectory,
    awaitWriteFinish: {
      stabilityThreshold: 100
    },
    ignored: `*.d.ts`
  }).on(`change`, async path => {
    if (path.endsWith(`.d.ts`)) return;
    const extension = extname(path);
    if (!supportedExtensions.includes(extension)) return;
    const scriptName = basename(path, extension);

    // if the path is still the same after getting just the base, it means it's directly in the source directory (global)
    if (path == basename(path)) {
      if (!pushEverything && !wildScriptUsers.size && !wildUserScripts.has(scriptName) && !scriptNamesToUsers.has(scriptName)) return;
      const scriptNamesToUsersToSkip = new DynamicMap(_scriptName => []);
      await Promise.all((await readdir(sourceDirectory, {
        withFileTypes: true
      })).map(async dirent => {
        if (!dirent.isDirectory()) return;
        for (const file of await readdir(resolve(sourceDirectory, dirent.name), {
          withFileTypes: true
        })) {
          if (!file.isFile()) continue;
          const fileExtension = extname(file.name);
          if (supportedExtensions.includes(fileExtension)) scriptNamesToUsersToSkip.get(basename(file.name, fileExtension)).push(dirent.name);
        }
      }));
      const usersToPushToSet = new Set();
      if (pushEverything || wildUserScripts.has(scriptName)) {
        for (const dirent of await readdir(resolve(sourceDirectory), {
          withFileTypes: true
        })) {
          if (dirent.isDirectory()) usersToPushToSet.add(dirent.name);
        }
        for (const dirent of await readdir(resolve(hackmudDirectory), {
          withFileTypes: true
        })) {
          if (dirent.isDirectory()) usersToPushToSet.add(dirent.name);else if (dirent.isFile() && dirent.name.endsWith(`.key`)) usersToPushToSet.add(dirent.name.slice(0, -4));
        }
        for (const users of scriptNamesToUsers.values()) {
          for (const user of users) usersToPushToSet.add(user);
        }
      }
      for (const user of wildScriptUsers) usersToPushToSet.add(user);
      for (const user of scriptNamesToUsers.get(scriptName)) usersToPushToSet.add(user);
      const usersToPushTo = [...usersToPushToSet].filter(user => !scriptNamesToUsersToSkip.has(user));
      if (!usersToPushTo.length) {
        onPush?.({
          file: path,
          users: [],
          minLength: 0,
          error: new Error(`no users to push to`)
        });
        return;
      }
      const uniqueID = Math.floor(Math.random() * 2 ** 52).toString(36).padStart(11, `0`);
      const filePath = resolve(sourceDirectory, path);
      let minifiedCode;
      try {
        ({
          script: minifiedCode
        } = await processScript(await readFile(filePath, {
          encoding: `utf-8`
        }), {
          minify,
          scriptUser: true,
          scriptName,
          uniqueID,
          filePath,
          mangleNames,
          forceQuineCheats
        }));
      } catch (error) {
        assert(error instanceof Error);
        onPush?.({
          file: path,
          users: [],
          minLength: 0,
          error
        });
        return;
      }
      await Promise.all(usersToPushTo.map(user => writeFilePersistent(resolve(hackmudDirectory, user, `scripts/${scriptName}.js`), minifiedCode.replace(new RegExp(`\\$${uniqueID}\\$SCRIPT_USER\\$`, `g`), user).replace(new RegExp(`\\$${uniqueID}\\$FULL_SCRIPT_NAME\\$`, `g`), `${user}.${scriptName}`))));
      onPush?.({
        file: path,
        users: usersToPushTo,
        minLength: countHackmudCharacters(minifiedCode),
        error: undefined
      });
      return;
    }
    const user = basename(resolve(path, `..`));
    if (!pushEverything && !wildScriptUsers.size && !wildUserScripts.has(scriptName) && !scriptNamesToUsers.get(scriptName).has(user)) return;
    const filePath = resolve(sourceDirectory, path);
    const sourceCode = await readFile(filePath, {
      encoding: `utf-8`
    });
    let script;
    try {
      ({
        script
      } = await processScript(sourceCode, {
        minify,
        scriptUser: user,
        scriptName,
        filePath,
        mangleNames,
        forceQuineCheats
      }));
    } catch (error) {
      assert(error instanceof Error);
      onPush?.({
        file: path,
        users: [],
        minLength: 0,
        error
      });
      return;
    }
    await writeFilePersistent(resolve(hackmudDirectory, user, `scripts`, `${scriptName}.js`), script);
    onPush?.({
      file: path,
      users: [user],
      minLength: countHackmudCharacters(script),
      error: undefined
    });
  });
  if (onReady) watcher.on(`ready`, onReady);
  if (!typeDeclarationPath_) return;
  let typeDeclarationPath = typeDeclarationPath_;

  /*
  	this currently works because the generated type declaration effectively
  	just connects things and doesn't actually read the scripts it's
  	generating type declarations for
  		if I ever change `generateTypings()` to actually read the scripts to
  	grab their types, this will need to change
  */

  const writeTypeDeclaration = async () => {
    const typeDeclaration = await generateTypeDeclaration(sourceDirectory, hackmudDirectory);
    try {
      await writeFile(typeDeclarationPath, typeDeclaration);
    } catch (error) {
      assert(error instanceof Error);
      if (!(error.code == `EISDIR`)) throw error;
      typeDeclarationPath = resolve(typeDeclarationPath, `player.d.ts`);
      await writeFile(typeDeclarationPath, typeDeclaration);
    }
  };
  await writeTypeDeclaration();
  watcher.on(`add`, writeTypeDeclaration);
  watcher.on(`unlink`, writeTypeDeclaration);
};

export { watch as default, watch };
